# -*- coding: utf-8 -*-
import urllib.parse

import jwt

from girder.api.rest import getApiUrl
from girder.exceptions import RestException
from girder.models.setting import Setting

from .base import ProviderBase
from ..settings import PluginSettings


class Google(ProviderBase):
    # https://developers.google.com/identity/protocols/OpenIDConnect
    _AUTH_URL = 'https://accounts.google.com/o/oauth2/v2/auth'
    _AUTH_SCOPES = ['profile', 'email']
    _TOKEN_URL = 'https://oauth2.googleapis.com/token'
    _DISCOVERY_URL = 'https://accounts.google.com/.well-known/openid-configuration'

    def getClientIdSetting(self):
        return Setting().get(PluginSettings.GOOGLE_CLIENT_ID)

    def getClientSecretSetting(self):
        return Setting().get(PluginSettings.GOOGLE_CLIENT_SECRET)

    @classmethod
    def getUrl(cls, state):
        clientId = Setting().get(PluginSettings.GOOGLE_CLIENT_ID)
        if not clientId:
            raise Exception('No Google client ID setting is present.')

        callbackUrl = '/'.join((getApiUrl(), 'oauth', 'google', 'callback'))

        query = urllib.parse.urlencode({
            'response_type': 'code',
            'access_type': 'online',
            'client_id': clientId,
            'redirect_uri': callbackUrl,
            'state': state,
            'scope': 'openid %s' % ' '.join(cls._AUTH_SCOPES)
            # Google supports a 'nonce', but does not require it, and Girder's particular checks
            # and invalidation of the 'state' token provides the same security guarantees
        })
        return '%s?%s' % (cls._AUTH_URL, query)

    def getToken(self, code):
        params = {
            'grant_type': 'authorization_code',
            'code': code,
            'client_id': self.clientId,
            'client_secret': self.clientSecret,
            'redirect_uri': self.redirectUri
        }
        resp = self._getJson(method='POST', url=self._TOKEN_URL, data=params)
        return resp

    def getUser(self, token):
        idToken = token['id_token']

        # Because the token came directly from Google's API, we don't need to verify it
        payload = jwt.decode(idToken, verify=False)

        oauthId = payload['sub']

        email = payload.get('email')
        if not email:
            raise RestException('This Google user has no available email address.', code=502)

        # The user's full name is in the payload, but is not split into first and last names

        discovery = self._getJson(method='GET', url=self._DISCOVERY_URL)
        userinfoUrl = discovery['userinfo_endpoint']

        userinfo = self._getJson(method='GET', url=userinfoUrl, headers={
            'Authorization': '%s %s' % (token['token_type'], token['access_token'])
        })
        firstName = userinfo.get('given_name', '')
        lastName = userinfo.get('family_name', '')

        user = self._createOrReuseUser(oauthId, email, firstName, lastName)
        return user



# -*- coding: utf-8 -*-
import cherrypy
import click
import errno
import fuse
import os
import stat
import sys
import threading
import time

import girder
from girder import events, logger, logprint
from girder.exceptions import AccessException, ValidationException
from girder.models.file import File
from girder.models.folder import Folder
from girder.models.item import Item
from girder.models.setting import Setting
from girder.settings import SettingKey
from girder.utility import config
from girder.utility.model_importer import ModelImporter
from girder.utility import path as path_util
from girder.utility.server import configureServer


class ServerFuse(fuse.Operations):
    """
    This class handles FUSE operations that are non-default.  It exposes the
    Girder resources via the resource path in a read-only manner.  It could be
    extended to expose metadata and other resources by extending the available
    paths.  Files can also be reached via a path shortcut of /file/<id>.
    """

    use_ns = True

    def __init__(self, stat=None):
        """
        Instantiate the operations class.  This sets up tracking for open
        files and file descriptor numbers (handles).

        :param stat: the results of an os.stat call which should be used as
            default values for files in the FUSE.  Files in the FUSE will have
            the same uid, gid, and atime.  If the resource lacks both an
            updated and a created time stamp, the ctime and mtime will also be
            taken from this.  If None, this defaults to the user of the Girder
            process's home directory,
        """
        super().__init__()
        if not stat:
            stat = os.stat(os.path.expanduser('~'))
        # we always set st_mode, st_size, st_ino, st_nlink, so we don't need
        # to track those.
        self._defaultStat = dict((key, getattr(stat, key)) for key in (
            'st_atime', 'st_ctime', 'st_gid', 'st_mtime', 'st_uid', 'st_blksize'))
        self.nextFH = 1
        self.openFiles = {}
        self.openFilesLock = threading.Lock()

    def __call__(self, op, path, *args, **kwargs):
        """
        Generically allow logging and error handling for any operation.

        :param op: operation to perform.
        :param path: path within the fuse (e.g., '', '/user', '/user/<name>',
            etc.).
        """
        logger.debug('-> %s %s %s', op, path, repr(args))
        ret = '[exception]'
        try:
            ret = getattr(self, op)(path, *args, **kwargs)
            return ret
        except Exception as e:
            # Log all exceptions and then reraise them
            if getattr(e, 'errno', None) in (errno.ENOENT, errno.EACCES):
                logger.debug('-- %s %r', op, e)
            else:
                logger.exception('-- %s', op)
            raise e
        finally:
            if op != 'read':
                logger.debug('<- %s %s', op, repr(ret))
            else:
                logger.debug('<- %s (length %d) %r', op, len(ret), ret[:16])

    def _getPath(self, path):
        """
        Given a fuse path, return the associated resource.

        :param path: path within the fuse.
        :returns: a Girder resource dictionary.
        """
        # If asked about a file in top level directory or the top directory,
        # return that it doesn't exist.  Other methods should handle '',
        # '/user', and 'collection' before calling this method.
        if '/' not in path.rstrip('/')[1:]:
            raise fuse.FuseOSError(errno.ENOENT)
        try:
            # We can't filter the resource, since that removes files'
            # assetstore information and users' size information.
            resource = path_util.lookUpPath(
                path.rstrip('/'), filter=False, force=True)
        except (path_util.NotFoundException, AccessException):
            raise fuse.FuseOSError(errno.ENOENT)
        except ValidationException:
            raise fuse.FuseOSError(errno.EROFS)
        except Exception:
            logger.exception('ServerFuse server internal error')
            raise fuse.FuseOSError(errno.EROFS)
        return resource   # {model, document}

    def _stat(self, doc, model):
        """
        Generate stat results for a resource.

        :param doc: the girder resource document.
        :param model: the girder model.
        :returns: the stat dictionary.
        """
        attr = self._defaultStat.copy()
        # We could specify distinct ino.  For instance, we could generate them
        # via a hash of the document ID (something like  int(hashlib.sha512(
        # str(doc['_id'])).hexdigest()[-8:], 16) ).  There doesn't seem to be
        # any measurable benefit of this, however, so we specify use_ino false
        # in the mount and set the value to -1 here.
        attr['st_ino'] = -1
        attr['st_nlink'] = 1
        if 'updated' in doc:
            attr['st_mtime'] = int(time.mktime(doc['updated'].timetuple()) * 1e9)
        elif 'created' in doc:
            attr['st_mtime'] = int(time.mktime(doc['created'].timetuple()) * 1e9)
        attr['st_ctime'] = attr['st_mtime']

        if model == 'file':
            attr['st_mode'] = 0o400 | stat.S_IFREG
            attr['st_size'] = doc.get('size', len(doc.get('linkUrl', '')))
        else:
            attr['st_mode'] = 0o500 | stat.S_IFDIR
            # Directories have zero size.  We could, instead, list the size
            # of all of their children via doc.get('size', 0), but that isn't
            # how most directories are reported.
            attr['st_size'] = 0
        return attr

    def _name(self, doc, model):
        """
        Return the name associated with a Girder resource.

        :param doc: the girder resource document.
        :param model: the girder model.
        :returns: the resource name as a text string.
        """
        name = path_util.getResourceName(model, doc)
        if isinstance(name, bytes):
            name = name.decode('utf8')
        name = path_util.encode(name)
        return name

    def _list(self, doc, model):
        """
        List the entries in a Girder user, collection, folder, or item.

        :param doc: the girder resource document.
        :param model: the girder model.
        :returns: a list of the names of resources within the specified
        document.
        """
        entries = []
        if model in ('collection', 'user', 'folder'):
            folderList = Folder().find({
                'parentId': doc['_id'],
                'parentCollection': model.lower()
            })
            for folder in folderList:
                entries.append(self._name(folder, 'folder'))
        if model == 'folder':
            for item in Folder().childItems(doc):
                entries.append(self._name(item, 'item'))
        elif model == 'item':
            for file in Item().childFiles(doc):
                entries.append(self._name(file, 'file'))
        return entries

    # We don't handle extended attributes or ioctl.
    getxattr = None
    listxattr = None
    ioctl = None

    def access(self, path, mode):
        """
        Try to load the resource associated with a path.  If we have permission
        to do so based on the current mode, report that access is allowed.
        Otherwise, an exception is raised.

        :param path: path within the fuse.
        :param mode: either F_OK to test if the resource exists, or a bitfield
            of R_OK, W_OK, and X_OK to test if read, write, and execute
            permissions are allowed.
        :returns: 0 if access is allowed.  An exception is raised if it is
            not.
        """
        if path.rstrip('/') in ('', '/user', '/collection'):
            return super().access(path, mode)
        # mode is either F_OK or a bitfield of R_OK, W_OK, X_OK
        return 0

    def create(self, path, mode):
        """
        This is a read-only system, so don't allow create.
        """
        raise fuse.FuseOSError(errno.EROFS)

    def flush(self, path, fh=None):
        """
        We may want to disallow flush, since this is a read-only system:
            raise fuse.FuseOSError(errno.EACCES)
        For now, always succeed.
        """
        return 0

    def getattr(self, path, fh=None):
        """
        Get the attributes dictionary of a path.

        :param path: path within the fuse.
        :param fh: an open file handle.  Ignored, since path is always
            specified.
        :returns: an attribute dictionary.
        """
        if path.rstrip('/') in ('', '/user', '/collection'):
            attr = self._defaultStat.copy()
            attr['st_mode'] = 0o500 | stat.S_IFDIR
            attr['st_size'] = 0
        else:
            resource = self._getPath(path)
            attr = self._stat(resource['document'], resource['model'])
        if attr.get('st_blksize') and attr.get('st_size'):
            attr['st_blocks'] = int(
                (attr['st_size'] + attr['st_blksize'] - 1) / attr['st_blksize'])
        return attr

    def read(self, path, size, offset, fh):
        """
        Read a block of bytes from a resource.

        :param path: path within the fuse.  Ignored, since the fh parameter
            must be valid.
        :param size: maximum number of bytes to read.  There may be less if
            this is near the end of the file.
        :param offset: the offset within the file to read.
        :param fh: an open file handle.
        :returns: a block of up to <size> bytes.
        """
        with self.openFilesLock:
            if fh not in self.openFiles:
                raise fuse.FuseOSError(errno.EBADF)
            info = self.openFiles[fh]
        with info['lock']:
            handle = info['handle']
            handle.seek(offset)
            return handle.read(size)

    def readdir(self, path, fh):
        """
        Get a list of names within a directory.

        :param path: path within the fuse.
        :param fh: an open file handle.  Ignored, since path is always
            specified.
        :returns: a list of names.  This always includes . and ..
        """
        path = path.rstrip('/')
        result = [u'.', u'..']
        if path == '':
            result.extend([u'collection', u'user'])
        elif path in ('/user', '/collection'):
            model = path[1:]
            docList = ModelImporter.model(model).find({}, sort=None)
            for doc in docList:
                result.append(self._name(doc, model))
        else:
            resource = self._getPath(path)
            result.extend(self._list(resource['document'], resource['model']))
        return result

    def open(self, path, flags):
        """
        Open a path and return a descriptor.

        :param path: path within the fuse.
        :param flags: a combination of O_* flags.  This will fail if it is not
            read only.
        :returns: a file descriptor.
        """
        resource = self._getPath(path)
        if resource['model'] != 'file':
            return super().open(path, flags)
        if flags & (os.O_APPEND | os.O_ASYNC | os.O_CREAT | os.O_DIRECTORY
                    | os.O_EXCL | os.O_RDWR | os.O_TRUNC | os.O_WRONLY):
            raise fuse.FuseOSError(errno.EROFS)
        info = {
            'path': path,
            'handle': File().open(resource['document']),
            'lock': threading.Lock(),
        }
        with self.openFilesLock:
            fh = self.nextFH
            self.nextFH += 1
            self.openFiles[fh] = info
        return fh

    def release(self, path, fh):
        """
        Release an open file handle.

        :param path: path within the fuse.
        :param fh: an open file handle.
        :returns: a file descriptor.
        """
        with self.openFilesLock:
            if fh in self.openFiles:
                with self.openFiles[fh]['lock']:
                    if 'handle' in self.openFiles[fh]:
                        self.openFiles[fh]['handle'].close()
                        del self.openFiles[fh]['handle']
                    del self.openFiles[fh]
            else:
                return super().release(path, fh)
        return 0

    def destroy(self, path):
        """
        Handle shutdown of the FUSE.

        :param path: always '/'.
        """
        Setting().unset(SettingKey.GIRDER_MOUNT_INFORMATION)
        events.trigger('server_fuse.destroy')
        return super().destroy(path)


class FUSELogError(fuse.FUSE):
    def __init__(self, operations, mountpoint, *args, **kwargs):
        """
        This wraps fuse.FUSE so that errors are logged rather than raising a
        RuntimeError exception.
        """
        try:
            logger.debug('Mounting %s\n' % mountpoint)
            super().__init__(operations, mountpoint, *args, **kwargs)
            logger.debug('Mounted %s\n' % mountpoint)
        except RuntimeError:
            logprint.error(
                'Failed to mount FUSE.  Does the mountpoint (%r) exist and is '
                'it empty?  Does the user have permission to create FUSE '
                'mounts?  It could be another FUSE mount issue, too.' % (
                    mountpoint, ))
            Setting().unset(SettingKey.GIRDER_MOUNT_INFORMATION)


def unmountServer(path, lazy=False, quiet=False):
    """
    Unmount a specified path, if possible.

    :param path: the path to unmount.
    :param lazy: True to pass the lazy flag to the unmount command.
    :returns: the return code of the unmount program (0 for success).  A
        non-zero code could mean that the unmount failed or was not needed.
    """
    # We only import these for the unmount command
    import shutil
    import subprocess

    if shutil.which('fusermount'):
        cmd = ['fusermount', '-u']
        if lazy:
            cmd.append('-z')
    else:
        cmd = ['umount']
        if lazy:
            cmd.append('-l')
    cmd.append(os.path.realpath(path))
    if quiet:
        with open(getattr(os, 'devnull', '/dev/null'), 'w') as devnull:
            result = subprocess.call(cmd, stdout=devnull, stderr=devnull)
    else:
        result = subprocess.call(cmd)
    return result


@click.command(
    'mount', short_help='Mount Girder files.',
    help='Mount Girder files via a read-only FUSE.  Specify the path for the mountpoint.')
@click.argument('path', type=click.Path(exists=True))
@click.option(
    '-d', '--database', default=cherrypy.config['database']['uri'],
    show_default=True,
    help='The database URI to connect to.  If this does not include a ://, '
         'the default database will be used.')
@click.option(
    '-o', '--options', 'fuseOptions', default=None,
    help='Comma separated list of additional FUSE mount options.  '
         'ro and use_ino cannot be overridden.')
@click.option(
    '-q', '--quiet', is_flag=True, default=False,
    help='Suppress Girder startup information or unmount output.')
@click.option(
    '-u', '--umount', '--unmount', 'unmount', is_flag=True, default=False,
    help='Unmount a mounted FUSE filesystem.')
@click.option(
    '-l', '-z', '--lazy', 'lazy', is_flag=True, default=False,
    help='Lazy unmount.')
@click.option('--plugins', default=None, help='Comma separated list of plugins to import.')
def main(path, database, fuseOptions, quiet, unmount, lazy, plugins):
    if unmount or lazy:
        result = unmountServer(path, lazy, quiet)
        sys.exit(result)
    mountServer(path=path, database=database, fuseOptions=fuseOptions,
                quiet=quiet, plugins=plugins)


def mountServer(path, database=None, fuseOptions=None, quiet=False, plugins=None):
    """
    Perform the mount.

    :param path: the mount location.
    :param database: a database connection URI, if it contains '://'.
        Otherwise, the default database is used.
    :param fuseOptions: a comma-separated string of options to pass to the FUSE
        mount.  A key without a value is taken as True.  Boolean values are
        case insensitive.  For instance, 'foreground' or 'foreground=True' will
        keep this program running until the SIGTERM or unmounted.
    :param quiet: if True, suppress Girder logs.
    :param plugins: an optional list of plugins to enable.  If None, use the
        plugins that are configured.
    """
    if quiet:
        curConfig = config.getConfig()
        curConfig.setdefault('logging', {})['log_quiet'] = True
        curConfig.setdefault('logging', {})['log_level'] = 'FATAL'
        girder._attachFileLogHandlers()
    if database and '://' in database:
        cherrypy.config['database']['uri'] = database
    if plugins is not None:
        plugins = plugins.split(',')
    webroot, appconf = configureServer(plugins=plugins)
    girder._setupCache()

    opClass = ServerFuse(stat=os.stat(path))
    options = {
        # By default, we run in the background so the mount command returns
        # immediately.  If we run in the foreground, a SIGTERM will shut it
        # down
        'foreground': False,
        # Cache files if their size and timestamp haven't changed.
        # This lets the OS buffer files efficiently.
        'auto_cache': True,
        # We aren't specifying our own inos
        'use_ino': False,
        # read-only file system
        'ro': True,
    }
    if sys.platform != 'darwin':
        # Automatically unmount when we try to mount again
        options['auto_unmount'] = True
    if fuseOptions:
        for opt in fuseOptions.split(','):
            if '=' in opt:
                key, value = opt.split('=', 1)
                value = (False if value.lower() == 'false' else
                         True if value.lower() == 'true' else value)
            else:
                key, value = opt, True
            if key in ('use_ino', 'ro', 'rw') and options.get(key) != value:
                logprint.warning('Ignoring the %s=%r option' % (key, value))
                continue
            options[key] = value
    Setting().set(SettingKey.GIRDER_MOUNT_INFORMATION,
                  {'path': path, 'mounttime': time.time()})
    FUSELogError(opClass, path, **options)


# You can add girder to the list of known filesystem types in linux.  For
# instance, if you create an executable file at /sbin/mount.girder that
# contains
#   #!/usr/bin/env bash
#   sudo -u <user that runs girder> girder mount -d "$@"
# then the command
#   mount -t girder <database uri> <path> -o <options>
# will work.  If you specify a string that doesn't contain :// as the database
# uri, then it will use the default girder configuration (e.g., use
# "mount -t girder girder <path>").
# If you have girder installed in a virtualenv, ifrequently prepending the
# virtualenv's bin directory to the path is enough to use it, so the
# mount.girder file becomes
#   #!/usr/bin/env bash
#   sudo -u <user that runs girder> bash -c 'PATH="<virtualenv
#       path>:$PATH" ${0} ${1+"$@"}' girder mount -d "$@"



import io
import itertools
import os
import re

import cherrypy
from cherrypy._cpreqbody import Part

from girder.api.rest import setResponseHeader, setContentDisposition
from girder.exceptions import GirderException, ValidationException, FilePathException
from girder.models.setting import Setting
from girder.settings import SettingKey
from girder.utility import progress, RequestBodyStream


class FileHandle:
    """
    This is the base class that is returned for the file-like API into
    Girder file objects. The ``open`` method of assetstore implementations
    is responsible for returning an instance of this class or one of its
    subclasses. This base class implementation is returned by the
    abstract assetstore adapter, and does not leverage any details of the
    assetstore implementations.

    These file handles are stateful, and therefore not safe for concurrent
    access. If used by multiple threads, mutexes should be used.

    :param file: The file object to which this file-like object corresponds.
    :type file: dict
    :param adapter: The assetstore adapter corresponding to this file.
    :type adapter: girder.utility.abstract_assetstore_adapter.AbstractAssetstoreAdapter
    """

    def __init__(self, file, adapter):
        self._file = file
        self._adapter = adapter
        self._pos = None
        # If a read is requested that is longer than the specified size, raise
        # an exception.  This prevents unbounded memory use.
        self._maximumReadSize = 16 * 1024 * 1024

        self.seek(0)

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()

    def read(self, size=None):
        """
        Read *size* bytes from the file data.

        :param size: The number of bytes to read from the current position. The
            actual number returned could be less than this if the end of the
            file is reached. An empty response indicates that the file has been
            completely consumed.  If None or negative, read to the end of the
            file.
        :type size: int
        :rtype: bytes
        """
        if size is None or size < 0:
            size = self._file['size'] - self._pos
        if size > self._maximumReadSize:
            raise GirderException('Read exceeds maximum allowed size.')
        data = io.BytesIO()
        length = 0
        for chunk in itertools.chain(self._prev, self._stream):
            chunkLen = len(chunk)

            if chunkLen == 0:
                break

            if length + chunkLen <= size:
                data.write(chunk)
                self._prev = []
                length += chunkLen

                if length == size:
                    break
            else:
                chunkLen = min(size - length, chunkLen)
                data.write(chunk[:chunkLen])
                self._prev = [chunk[chunkLen:]]
                length += chunkLen
                break

        self._pos += length
        return data.getvalue()

    def tell(self):
        return self._pos

    def seek(self, offset, whence=os.SEEK_SET):
        oldPos = self._pos

        if whence == os.SEEK_SET:
            self._pos = offset
        elif whence == os.SEEK_CUR:
            self._pos += offset
        elif whence == os.SEEK_END:
            self._pos = max(self._file['size'] + offset, 0)

        if self._pos != oldPos:
            self._prev = []
            self._stream = self._adapter.downloadFile(self._file, offset=self._pos, headers=False)()

    def close(self):
        pass

    def seekable(self):
        return True


class AbstractAssetstoreAdapter:
    """
    This defines the interface to be used by all assetstore adapters.
    """

    def __init__(self, assetstore):
        self.assetstore = assetstore

    @staticmethod
    def validateInfo(doc):
        """
        Adapters may implement this if they need to perform any validation
        steps whenever the assetstore info is saved to the database. It should
        return the document with any necessary alterations in the success case,
        or throw an exception if validation fails.
        """
        return doc

    @staticmethod
    def fileIndexFields():
        """
        Default behavior is that no additional file fields need to be indexed
        within the database.
        """
        return []

    def capacityInfo(self):
        """
        Assetstore types that are able to report how much free and/or total
        capacity they have should override this method. Default behavior is to
        report both quantities as unknown.

        :returns: A dict with 'free' and 'total' keys whose values are
                  either bytes (ints) or None for an unknown quantity.
        """
        return {
            'free': None,
            'total': None
        }

    def initUpload(self, upload):
        """
        This must be called before any chunks are uploaded to do any
        additional behavior and optionally augment the upload document. The
        method must return the upload document. Default behavior is to
        simply return the upload document unmodified.

        :param upload: The upload document to optionally augment.
        :type upload: dict
        """
        return upload

    def uploadChunk(self, upload, chunk):
        """
        Call this method to process each chunk of an upload.

        :param upload: The upload document to update.
        :type upload: dict
        :param chunk: The file object representing the chunk that was uploaded.
        :type chunk: file
        :returns: Must return the upload document with any optional changes.
        """
        raise NotImplementedError('Must override processChunk in %s.' %
                                  self.__class__.__name__)

    def finalizeUpload(self, upload, file):
        """
        Call this once the last chunk has been processed. This method does not
        need to delete the upload document as that will be deleted by the
        caller afterward. This method may augment the File document, and must
        return the File document.

        :param upload: The upload document.
        :type upload: dict
        :param file: The file document that was created.
        :type file: dict
        :returns: The file document with optional modifications.
        """
        return file

    def requestOffset(self, upload):
        """
        Request the offset for resuming an interrupted upload. Default behavior
        simply returns the 'received' field of the upload document. This method
        exists because in some cases, such as when the server crashes, it's
        possible that the received field is not accurate, so adapters may
        implement this to provide the actual next byte required.
        """
        return upload['received']

    def deleteFile(self, file):
        """
        This is called when a File is deleted to allow the adapter to remove
        the data from within the assetstore. This method should not modify
        or delete the file object, as the caller will delete it afterward.

        :param file: The File document about to be deleted.
        :type file: dict
        """
        raise NotImplementedError('Must override deleteFile in %s.' %
                                  self.__class__.__name__)

    def shouldImportFile(self, path, params):
        """
        This is a helper used during the import process to determine if a file located at
        the specified path should be imported, based on the request parameters. Exclusion
        takes precedence over inclusion.

        :param path: The path of the file.
        :type path: str
        :param params: The request parameters.
        :type params: dict
        :rtype: bool
        """
        include = params.get('fileIncludeRegex')
        exclude = params.get('fileExcludeRegex')

        fname = os.path.basename(path)

        if exclude and re.match(exclude, fname):
            return False

        if include:
            return re.match(include, fname)

        return True

    def downloadFile(self, file, offset=0, headers=True, endByte=None,
                     contentDisposition=None, extraParameters=None, **kwargs):
        """
        This method is in charge of returning a value to the RESTful endpoint
        that can be used to download the file. This should either return a
        generator function that yields the bytes of the file (which will stream
        the file directly), or modify the response headers and raise a
        `cherrypy.HTTPRedirect`.

        :param file: The file document being downloaded.
        :type file: dict
        :param offset: Offset in bytes to start the download at.
        :type offset: int
        :param headers: Flag for whether headers should be sent on the response.
        :type headers: bool
        :param endByte: Final byte to download. If ``None``, downloads to the
            end of the file.
        :type endByte: int or None
        :param contentDisposition: Value for Content-Disposition response
            header disposition-type value.
        :type contentDisposition: str or None
        :type extraParameters: str or None
        """
        raise NotImplementedError('Must override downloadFile in %s.' %
                                  self.__class__.__name__)

    def findInvalidFiles(self, progress=progress.noProgress, filters=None,
                         checkSize=True, **kwargs):
        """
        Finds and yields any invalid files in the assetstore. It is left to
        the caller to decide what to do with them.

        :param progress: Pass a progress context to record progress.
        :type progress: :py:class:`girder.utility.progress.ProgressContext`
        :param filters: Additional query dictionary to restrict the search for
            files. There is no need to set the ``assetstoreId`` in the filters,
            since that is done automatically.
        :type filters: dict or None
        :param checkSize: Whether to make sure the size of the underlying
            data matches the size of the file.
        :type checkSize: bool
        """
        raise NotImplementedError('Must override findInvalidFiles in %s.' %
                                  self.__class__.__name__)

    def copyFile(self, srcFile, destFile):
        """
        This method copies the necessary fields and data so that the
        destination file contains the same data as the source file.

        :param srcFile: The original File document.
        :type srcFile: dict
        :param destFile: The File which should have the data copied to it.
        :type destFile: dict
        :returns: A dict with the destination file.
        """
        return destFile

    def getChunkSize(self, chunk):
        """
        Given a chunk that is either a file-like object or a string, attempt to
        determine its length.  If it is a file-like object, then this relies on
        being able to use fstat.

        :param chunk: the chunk to get the size of
        :type chunk: a file-like object or a string
        :returns: the length of the chunk if known, or None.
        """
        if isinstance(chunk, (io.BytesIO, RequestBodyStream, Part)):
            return
        elif hasattr(chunk, 'fileno'):
            return os.fstat(chunk.fileno()).st_size
        elif isinstance(chunk, str):
            return len(chunk.encode('utf8'))
        else:
            return len(chunk)

    def setContentHeaders(self, file, offset, endByte, contentDisposition=None):
        """
        Sets the Content-Length, Content-Disposition, Content-Type, and also
        the Content-Range header if this is a partial download.

        :param file: The file being downloaded.
        :param offset: The start byte of the download.
        :type offset: int
        :param endByte: The end byte of the download (non-inclusive).
        :type endByte: int
        :param contentDisposition: Content-Disposition response header
            disposition-type value, if None, Content-Disposition will
            be set to 'attachment; filename=$filename'.
        :type contentDisposition: str or None
        """
        isRangeRequest = cherrypy.request.headers.get('Range')
        setResponseHeader(
            'Content-Type',
            file.get('mimeType') or 'application/octet-stream')
        setContentDisposition(file['name'], contentDisposition or 'attachment')
        setResponseHeader('Content-Length', max(endByte - offset, 0))

        if (offset or endByte < file['size'] or isRangeRequest) and file['size']:
            setResponseHeader(
                'Content-Range',
                'bytes %d-%d/%d' % (offset, endByte - 1, file['size']))

    def checkUploadSize(self, upload, chunkSize):
        """
        Check if the upload is valid based on the chunk size.  If this
        raises an exception, then the caller should clean up and reraise the
        exception.

        :param upload: the dictionary of upload information.  The received and
                       size values are used.
        :param chunkSize: the chunk size that needs to be validated.
        :type chunkSize: a non-negative integer or None if unknown.
        """
        if 'received' not in upload or 'size' not in upload:
            return
        if chunkSize is None:
            return
        if upload['received'] + chunkSize > upload['size']:
            raise ValidationException('Received too many bytes.')
        if (upload['received'] + chunkSize != upload['size']
                and chunkSize < Setting().get(SettingKey.UPLOAD_MINIMUM_CHUNK_SIZE)):
            raise ValidationException('Chunk is smaller than the minimum size.')

    def cancelUpload(self, upload):
        """
        This is called when an upload has been begun and it should be
        abandoned.  It must clean up temporary files, chunks, or whatever other
        information the assetstore contains.
        """
        raise NotImplementedError('Must override cancelUpload in %s.' %
                                  self.__class__.__name__)

    def untrackedUploads(self, knownUploads=(), delete=False):
        """
        List and optionally discard uploads that are in the assetstore but not
        in the known list.

        :param knownUploads: a list of upload dictionaries of all known
                             incomplete uploads.
        :type knownUploads: list
        :param delete: if True, delete any unknown uploads.
        :type delete: bool
        :returns: a list of unknown uploads.
        """
        return ()

    def importData(self, parent, parentType, params, progress, user, **kwargs):
        """
        Assetstores that are capable of importing pre-existing data from the
        underlying storage medium can implement this method.

        :param parent: The parent object to import into.
        :param parentType: The model type of the parent object (folder, user,
            or collection).
        :type parentType: str
        :param params: Additional parameters required for the import process.
            Typically includes an importPath field representing a root path
            on the underlying storage medium.
        :type params: dict
        :param progress: Object on which to record progress if possible.
        :type progress: :py:class:`girder.utility.progress.ProgressContext`
        :param user: The Girder user performing the import.
        :type user: dict or None
        """
        raise NotImplementedError(
            'The %s assetstore type does not support importing existing data.'
            % self.__class__.__name__)

    def fileUpdated(self, file):
        """
        This is called when the file document has been changed. Any assetstore
        implementation that needs to do anything when the file document changes
        should override this method.

        :param file: The updated file document.
        :type file: dict
        """
        pass

    def open(self, file):
        """
        Exposes a Girder file as a python file-like object. At the
        moment, this is a read-only interface, the equivalent of opening a
        system file with 'rb' mode.

        :param file: A Girder file document.
        :type file: dict
        :return: A file-like object containing the bytes of the file.
        :rtype: FileHandle
        """
        return FileHandle(file, self)

    def getLocalFilePath(self, file):
        """
        If an assetstore adapter supports it, return a path to the file on the
        local file system.  Otherwise, raise an exception.

        :param file: The file document.
        :type file: dict
        :returns: a local path to the file.
        :rtype: str
        """
        raise FilePathException('This assetstore does not expose file paths')



# -*- coding: utf-8 -*-
import datetime
import json
import os

from tests import base
from girder.constants import AssetstoreType
from girder.exceptions import ValidationException
from girder.models.assetstore import Assetstore
from girder.models.collection import Collection
from girder.models.folder import Folder
from girder.models.setting import Setting
from girder.models.user import User
from girder.settings import SettingKey
from girder.utility.system import formatSize

from girder_user_quota.settings import PluginSettings


def setUpModule():
    base.enabledPlugins.append('user_quota')
    base.startServer()


def tearDownModule():
    base.stopServer()
    base.dropAllTestDatabases()


class QuotaTestCase(base.TestCase):
    def setUp(self):
        super().setUp()
        admin = {
            'email': 'admin@girder.test',
            'login': 'adminlogin',
            'firstName': 'Admin',
            'lastName': 'Last',
            'password': 'adminpassword',
            'admin': True
        }
        self.admin = User().createUser(**admin)
        user = {
            'email': 'good@girder.test',
            'login': 'goodlogin',
            'firstName': 'First',
            'lastName': 'Last',
            'password': 'goodpassword',
            'admin': False
        }
        self.user = User().createUser(**user)
        coll = {
            'name': 'Test Collection',
            'description': 'The description',
            'public': True,
            'creator': self.admin
        }
        self.collection = Collection().createCollection(**coll)
        Folder().createFolder(
            parent=self.collection, parentType='collection', name='Public',
            public=True, creator=self.admin)

    def _uploadFile(self, name, parent, parentType='folder', size=1024,
                    error=None, partial=False, validationError=None):
        """
        Upload a random file to an item.

        :param name: name of the file.
        :param parent: parent document to upload the file to.
        :param parentType: type of parent to upload to.
        :param size: size of the file to upload
        :param error: if set, expect this error.
        :param partial: if set, return before uploading the data.  This returns
                        a kwargs for self.request so that the caller can
                        finish the upload later.
        :returns: file: the created file object
        """
        if parentType != 'file':
            resp = self.request(
                path='/file', method='POST', user=self.admin, params={
                    'parentType': parentType,
                    'parentId': parent['_id'],
                    'name': name,
                    'size': size,
                    'mimeType': 'application/octet-stream'
                }, exception=error is not None)
        else:
            resp = self.request(
                path='/file/%s/contents' % str(parent['_id']), method='PUT',
                user=self.admin, params={
                    'size': size,
                }, exception=error is not None)
        if error:
            self.assertStatus(resp, 500)
            self.assertEqual(resp.json['type'], 'girder')
            self.assertEqual(resp.json['message'][:len(error)], error)
            return None
        elif validationError:
            self.assertStatus(resp, 400)
            self.assertEqual(resp.json['type'], 'validation')
            self.assertEqual(resp.json['message'][:len(validationError)], validationError)
            return None
        # We don't create the contents until after we check for the first
        # error.  This means that we can try to upload huge files without
        # allocating the space for them
        contents = os.urandom(size)
        self.assertStatusOk(resp)
        upload = resp.json
        if partial:
            body = contents[:-1]
        else:
            body = contents

        resp = self.request(path='/file/chunk', method='POST', user=self.admin, body=body, params={
            'uploadId': upload['_id']
        }, type='application/octet-stream')
        self.assertStatusOk(resp)
        if partial:
            return {
                'path': '/file/chunk',
                'method': 'POST',
                'user': self.admin,
                'body': contents[-1:],
                'type': 'application/octet-stream',
                'params': {
                    'offset': len(contents) - 1,
                    'uploadId': upload['_id']
                }
            }
        return resp.json

    def _setPolicy(self, policy, model, resource, user, error=None, results=None):
        """
        Set the quota or assetstore policy and check that it was set.

        :param policy: dictionary of values to set.
        :param model: either 'user' or 'collection'.
        :param resource: the document for the resource to test.
        :param user: user to use for authorization.
        :param error: if set, this is a substring expected in an error message.
        :param results: if not specified, we expect policy to match the set
                        results.  Otherwise, these are the expected results.
        """
        if isinstance(policy, dict):
            policyJSON = json.dumps(policy)
        else:
            policyJSON = policy
        path = '/%s/%s/quota' % (model, resource['_id'])
        resp = self.request(path=path, method='PUT', user=user,
                            params={'policy': policyJSON})
        if error:
            self.assertStatus(resp, 400)
            self.assertIn(error, resp.json['message'])
            return
        self.assertStatusOk(resp)
        resp = self.request(path=path, method='GET', user=user)
        self.assertStatusOk(resp)
        currentPolicy = resp.json['quota']
        if not results:
            results = policy
        for key in results:
            if results[key] or results[key] == 0 or results[key] is False:
                self.assertEqual(currentPolicy[key], results[key])
            else:
                self.assertEqual(currentPolicy[key], None)

    def _setQuotaDefault(self, model, value, testVal='__NOCHECK__',
                         error=None):
        """
        Set the default quota for a particular model.

        :param model: either 'user' or 'collection'.
        :param value: the value to set.  Either None or a positive integer.
        :param testVal: if not __NOCHECK__, test the current value to see if it
                        matches this.
        :param error: if set, this is a substring expected in an error message.
        """
        if model == 'user':
            key = PluginSettings.DEFAULT_USER_QUOTA
        elif model == 'collection':
            key = PluginSettings.DEFAULT_COLLECTION_QUOTA
        try:
            Setting().set(key, value)
        except ValidationException as err:
            if not error:
                raise
            if error not in err.args[0]:
                raise
            return
        if testVal != '__NOCHECK__':
            newVal = Setting().get(key)
            self.assertEqual(newVal, testVal)

    def _testAssetstores(self, model, resource, user):
        """
        Test assetstore policies for a specified resource.

        :param model: either 'user' or 'collection'.
        :param resource: the document for the resource to test.
        :param user: user to use for authorization.
        """
        resp = self.request(path='/folder', method='GET', user=user,
                            params={'parentType': model,
                                    'parentId': resource['_id']})
        self.assertStatusOk(resp)
        self.assertGreaterEqual(len(resp.json), 1)
        folder = resp.json[0]
        curAssetstoreId = str(self.currentAssetstore['_id'])
        altAssetstoreId = str(self.alternateAssetstore['_id'])
        badAssetstoreId = str(self.brokenAssetstore['_id'])
        invalidAssetstoreId = str(folder['_id'])
        file = self._uploadFile('Upload to Current', folder)
        self.assertEqual(file['assetstoreId'], curAssetstoreId)
        # A simple preferredASsetstore policy should work
        self._setPolicy({'preferredAssetstore': altAssetstoreId},
                        model, resource, user)
        file = self._uploadFile('Upload to Alt', folder)
        self.assertEqual(file['assetstoreId'], altAssetstoreId)
        # Uploading to an invalid preferredAssetstore policy should send the
        # data to the current asssetstore
        self._setPolicy({'preferredAssetstore': invalidAssetstoreId},
                        model, resource, user)
        file = self._uploadFile('Upload to Invalid', folder)
        self.assertEqual(file['assetstoreId'], curAssetstoreId)
        # Uploading to an unreachable preferredAssetstore policy should send
        # the data to the current asssetstore
        self._setPolicy({'preferredAssetstore': badAssetstoreId},
                        model, resource, user)
        file = self._uploadFile('Upload to Bad', folder)
        self.assertEqual(file['assetstoreId'], curAssetstoreId)
        # We can specify a fallback
        self._setPolicy({'fallbackAssetstore': altAssetstoreId},
                        model, resource, user)
        file = self._uploadFile('Upload to Bad with fallback', folder)
        self.assertEqual(file['assetstoreId'], altAssetstoreId)
        # Or say there shouldn't be a fallback
        self._setPolicy({'fallbackAssetstore': 'none'},
                        model, resource, user)
        self._uploadFile('Upload to Bad with no fallback', folder,
                         error='Required assetstore is unavailable')
        # if we clear the preferred policy, the fallback shouldn't matter.
        self._setPolicy({'preferredAssetstore': ''},
                        model, resource, user)
        file = self._uploadFile('Upload with not preferred', folder)
        self.assertEqual(file['assetstoreId'], curAssetstoreId)

    def _testQuota(self, model, resource, user):
        """
        Test quota policies for a specified resource.

        :param model: either 'user' or 'collection'.
        :param resource: the document for the resource to test.
        :param user: user to use for authorization.
        """
        Setting().set(SettingKey.UPLOAD_MINIMUM_CHUNK_SIZE, 0)
        resp = self.request(path='/folder', method='GET', user=user,
                            params={'parentType': model,
                                    'parentId': resource['_id']})
        self.assertStatusOk(resp)
        self.assertGreaterEqual(len(resp.json), 1)
        folder = resp.json[0]
        # Set 0 b quota so nothing can be uploaded
        self._setQuotaDefault(model, 0)
        # Upload one file, it should fail
        self._uploadFile(
            'File too large',
            folder,
            size=1024,
            validationError='Upload would exceed file storage quota',
        )
        # Upload a tiny file, it should also fail
        self._uploadFile(
            'File too large',
            folder,
            size=1,
            validationError='Upload would exceed file storage quota',
        )
        # Set quota to None so anything can be uploaded
        self._setQuotaDefault(model, None)
        # Set a 0 b policy so nothing can be uploaded
        self._setPolicy({'fileSizeQuota': 0, 'useQuotaDefault': False},
                        model, resource, user)
        # Upload one file, it should fail
        self._uploadFile(
            'File too large',
            folder,
            size=1024,
            validationError='Upload would exceed file storage quota',
        )
        # Upload a tiny file, it should also fail
        self._uploadFile(
            'File too large',
            folder,
            size=1,
            validationError='Upload would exceed file storage quota',
        )
        # Reset the policy to use the default so anything can be uploaded
        self._setPolicy({'useQuotaDefault': True}, model, resource, user)
        # Upload one file so that there is some use
        self._uploadFile('First upload', folder, size=1024)
        # Set a policy limiting things to 4 kb, then a 4 kb file should fail
        self._setPolicy({'fileSizeQuota': 4096, 'useQuotaDefault': False},
                        model, resource, user)
        self._uploadFile('File too large', folder, size=4096,
                         validationError='Upload would exceed file storage quota')
        # But a 2 kb file will succeed
        file = self._uploadFile('Second upload', folder, size=2048)
        # And a second 2 kb file will fail
        self._uploadFile('File too large', folder, size=2048,
                         validationError='Upload would exceed file storage quota')
        # If we start uploading two files, only one should complete
        file1kwargs = self._uploadFile('First partial', folder, size=768,
                                       partial=True)
        file2kwargs = self._uploadFile('Second partial', folder, size=768,
                                       partial=True)
        resp = self.request(**file1kwargs)
        self.assertStatusOk(resp)
        try:
            resp = self.request(**file2kwargs)
            self.assertStatus(resp, 400)
        except AssertionError as exc:
            self.assertTrue('Upload exceeded' in exc.args[0])
        # Shrink the quota to smaller than all of our files.  Replacing an
        # existing file should still work, though
        self._setPolicy({'fileSizeQuota': 2048}, model, resource, user)
        self._uploadFile('Second upload', file, 'file', size=1536)
        # Now test again using default quotas.  We have 1024+1536+768 = 3328
        # bytes currently uploaded
        self._setPolicy({'useQuotaDefault': True}, model, resource, user)
        # Upload should now be unlimited, so anything will work
        self._uploadFile('Fourth upload', folder, size=1792)
        # Set a policy limiting things to 8 kb, then an additional 4 kb file
        # should fail
        self._setQuotaDefault(model, 8192)
        self._uploadFile('File too large', folder, size=4096,
                         validationError='Upload would exceed file storage quota')
        # But a 2 kb file will succeed
        file = self._uploadFile('Fifth upload', folder, size=2048)
        # And a second 2 kb file will fail
        self._uploadFile('File too large', folder, size=2048,
                         validationError='Upload would exceed file storage quota')
        # Set a policy with a large quota to test using NumberLong in the
        # mongo settings.
        self._setQuotaDefault(model, 5 * 1024**3)
        # A small file should now upload
        file = self._uploadFile('Six upload', folder, size=2048)
        # But a huge one will fail
        self._uploadFile('File too large', folder, size=6 * 1024**3,
                         validationError='Upload would exceed file storage quota')

    def testAssetstorePolicy(self):
        """
        Test assetstore policies for a user and a collection.
        """
        # We want three assetstores for testing, one of which is unreachable.
        # We already have one, which is the current assetstore.
        base.dropGridFSDatabase('girder_test_user_quota_assetstore')
        params = {
            'name': 'Non-current Store',
            'type': AssetstoreType.GRIDFS,
            'db': 'girder_test_user_quota_assetstore'
        }
        resp = self.request(path='/assetstore', method='POST', user=self.admin,
                            params=params)
        self.assertStatusOk(resp)

        # Create a broken assetstore. (Must bypass validation since it should
        # not let us create an assetstore in a broken state).
        Assetstore().save({
            'name': 'Broken Store',
            'type': AssetstoreType.FILESYSTEM,
            'root': '/dev/null',
            'created': datetime.datetime.utcnow()
        }, validate=False)

        # Now get the assetstores and save their ids for later
        resp = self.request(path='/assetstore', method='GET', user=self.admin,
                            params={'sort': 'created'})
        self.assertStatusOk(resp)
        assetstores = resp.json
        self.assertEqual(len(assetstores), 3)
        self.currentAssetstore = assetstores[0]
        self.alternateAssetstore = assetstores[1]
        self.brokenAssetstore = assetstores[2]
        self._testAssetstores('user', self.user, self.admin)
        self._testAssetstores('collection', self.collection, self.admin)

    def testQuotaPolicy(self):
        """
        Test quota policies for a user and a collection.
        """
        self._testQuota('user', self.user, self.admin)
        self._testQuota('collection', self.collection, self.admin)

    def testPolicySettings(self):
        """
        Test validation of policy settings.
        """
        # We have to send a json dictionary
        # We can only set certain keys
        self._setPolicy('this is not json',
                        'user', self.user, self.admin,
                        error='Parameter policy must be valid JSON.')
        self._setPolicy(json.dumps(['this is not a dictionary']),
                        'user', self.user, self.admin,
                        error='Parameter policy must be a JSON object.')
        # We can only set certain keys
        self._setPolicy({'notAKey': 'notAValue'},
                        'user', self.user, self.admin,
                        error='not a valid quota policy key')
        # We need to pass None, current, or an ObjectId to preferredAssetstore
        self._setPolicy({'preferredAssetstore': 'current'},
                        'user', self.user, self.admin,
                        results={'preferredAssetstore': None})
        self._setPolicy({'preferredAssetstore': 'not an id'},
                        'user', self.user, self.admin,
                        error='Invalid preferredAssetstore')
        self._setPolicy({'preferredAssetstore': None},
                        'user', self.user, self.admin)
        # fallbackAssetstore can also take 'none'
        self._setPolicy({'fallbackAssetstore': 'current'},
                        'user', self.user, self.admin,
                        results={'fallbackAssetstore': None})
        self._setPolicy({'fallbackAssetstore': 'none'},
                        'user', self.user, self.admin)
        self._setPolicy({'fallbackAssetstore': 'not an id'},
                        'user', self.user, self.admin,
                        error='Invalid fallbackAssetstore')
        self._setPolicy({'fallbackAssetstore': None},
                        'user', self.user, self.admin)
        # fileSizeQuota can be None, blank, 0, or a positive integer
        self._setPolicy({'fileSizeQuota': 0},
                        'user', self.user, self.admin,
                        results={'fileSizeQuota': 0})
        self._setPolicy({'fileSizeQuota': '00'},
                        'user', self.user, self.admin,
                        results={'fileSizeQuota': 0})
        self._setPolicy({'fileSizeQuota': ''},
                        'user', self.user, self.admin,
                        results={'fileSizeQuota': None})
        self._setPolicy({'fileSizeQuota': 'not an integer'},
                        'user', self.user, self.admin,
                        error='Invalid quota')
        self._setPolicy({'fileSizeQuota': -1},
                        'user', self.user, self.admin,
                        error='Invalid quota')
        self._setPolicy({'fileSizeQuota': None},
                        'user', self.user, self.admin)
        # useDefaultQuota can be True, False, None, yes, no, 0, 1.
        valdict = {None: True, True: True, 'true': True, 1: True, 'True': True,
                   False: False, 'false': False, 0: False, 'False': False}
        for val in valdict:
            self._setPolicy({'useQuotaDefault': val},
                            'user', self.user, self.admin,
                            results={'useQuotaDefault': valdict[val]})
        self._setPolicy({'useQuotaDefault': 'not_a_boolean'}, 'user',
                        self.user, self.admin, error='Invalid useQuotaDefault')
        # the resource default values behave like fileSizeQuota
        self._setQuotaDefault('user', 0, 0)
        self._setQuotaDefault('user', '00', 0)
        self._setQuotaDefault('user', '', None)
        self._setQuotaDefault('user', 'not an integer', error='Invalid quota')
        self._setQuotaDefault('user', -1, error='Invalid quota')
        self._setQuotaDefault('user', None)

    def testFormatSize(self):
        """
        Test the formatSize function
        """
        testList = [
            (0, '0 B'),
            (1000, '1000 B'),
            (10000, '10000 B'),
            (20000, '19.53 kB'),
            (200000, '195.3 kB'),
            (2000000, '1.907 MB'),
            ]
        for testItem in testList:
            self.assertEqual(testItem[1], formatSize(testItem[0]))



# -*- coding: utf-8 -*-
import cherrypy
import os

import girder
from girder.constants import PACKAGE_DIR


def _mergeConfig(filename):
    """
    Load `filename` into the cherrypy config.
    Also, handle global options by putting them in the root.
    """
    cherrypy._cpconfig.merge(cherrypy.config, filename)
    # When in Sphinx, cherrypy may be mocked and returning None
    global_config = cherrypy.config.pop('global', {}) or {}

    for option, value in global_config.items():
        cherrypy.config[option] = value


def _loadConfigsByPrecedent():
    """
    Load configuration in reverse order of precedent.
    """
    # TODO: Deprecated, remove in a later version
    def _printConfigurationWarning():
        girder.logprint.warning(
            'Detected girder.local.cfg, this location is no longer supported.\n'
            'For supported locations, see '
            'https://girder.readthedocs.io/en/stable/configuration.html#configuration')

    if os.path.exists(os.path.join(PACKAGE_DIR, 'conf', 'girder.local.cfg')):
        # This can't use logprint since configuration is loaded before initialization.
        # Note this also won't be displayed when starting other services that don't start a CherryPy
        # server such as girder mount or girder sftpd.
        cherrypy.engine.subscribe('start', _printConfigurationWarning)

    configPaths = [os.path.join(PACKAGE_DIR, 'conf', 'girder.dist.cfg'),
                   os.path.join('/etc', 'girder.cfg'),
                   os.path.join(os.path.expanduser('~'), '.girder', 'girder.cfg')]

    if 'GIRDER_CONFIG' in os.environ:
        configPaths.append(os.environ['GIRDER_CONFIG'])

    for curConfigPath in configPaths:
        if os.path.exists(curConfigPath):
            _mergeConfig(curConfigPath)


def loadConfig():

    _loadConfigsByPrecedent()

    if 'GIRDER_PORT' in os.environ:
        port = int(os.environ['GIRDER_PORT'])
        cherrypy.config['server.socket_port'] = port

    if 'GIRDER_MONGO_URI' in os.environ:
        if 'database' not in cherrypy.config:
            cherrypy.config['database'] = {}
        cherrypy.config['database']['uri'] = os.getenv('GIRDER_MONGO_URI')

    if 'GIRDER_TEST_DB' in os.environ:
        cherrypy.config['database']['uri'] =\
            os.environ['GIRDER_TEST_DB'].replace('.', '_')


def getConfig():
    if 'database' not in cherrypy.config:
        loadConfig()
    # When in Sphinx, cherrypy may be mocked and returning None
    return cherrypy.config or {}


def getServerMode():
    return getConfig()['server']['mode']



# -*- coding: utf-8 -*-
import datetime

from .model_base import AccessControlledModel
from girder.constants import AccessType, TokenScope
from girder.exceptions import ValidationException
from girder.settings import SettingKey
from girder.utility import genToken


class ApiKey(AccessControlledModel):
    """
    This model represents API keys corresponding to users.
    """

    def initialize(self):
        self.name = 'api_key'
        self.ensureIndices(('userId', 'key'))

        self.exposeFields(level=AccessType.READ, fields={
            '_id', 'active', 'created', 'key', 'lastUse', 'name', 'scope',
            'tokenDuration', 'userId'
        })

    def validate(self, doc):
        from .token import Token
        from .user import User

        if doc['tokenDuration']:
            doc['tokenDuration'] = float(doc['tokenDuration'])
        else:
            doc['tokenDuration'] = None

        doc['name'] = doc['name'].strip()
        doc['active'] = bool(doc.get('active', True))

        if doc['scope'] is not None:
            if not isinstance(doc['scope'], (list, tuple)):
                raise ValidationException('Scope must be a list, or None.')
            if not doc['scope']:
                raise ValidationException('Custom scope list must not be empty.')

            # Ensure only registered scopes are being set
            admin = User().load(doc['userId'], force=True)['admin']
            scopes = TokenScope.scopeIds(admin)
            unknownScopes = set(doc['scope']) - scopes
            if unknownScopes:
                raise ValidationException('Invalid scopes: %s.' % ','.join(unknownScopes))

        # Deactivating an already existing token
        if '_id' in doc and not doc['active']:
            Token().clearForApiKey(doc)

        return doc

    def remove(self, doc):
        # Clear tokens corresponding to this API key.
        from .token import Token
        Token().clearForApiKey(doc)
        super().remove(doc)

    def list(self, user, limit=0, offset=0, sort=None):
        """
        List API keys for a given user.

        :param user: The user whose keys to list.
        :type user: dict
        :param limit: Result limit.
        :param offset: Result offset.
        :param sort: The sort structure to pass to pymongo.
        :rtype: iterable of API keys for the user.
        """
        return self.find({
            'userId': user['_id']
        }, limit=limit, offset=offset, sort=sort)

    def createApiKey(self, user, name, scope=None, days=None, active=True):
        """
        Create a new API key for a user.

        :param user: The user who owns the API key.
        :type user: dict
        :param name: A human readable name for the API key
        :param days: The lifespan of the session in days. If not passed, uses
            the database setting for cookie lifetime. Note that this is a
            maximum duration; clients may request tokens with a shorter lifetime
            than this value.
        :type days: float or int
        :param scope: Scope or list of scopes this API key grants. By default,
            will grant tokens provided full access on behalf of the user.
        :type scope: str, list of str, or set of str
        :param active: Whether this key is active.
        :returns: The API key document that was created.
        """
        apiKey = {
            'created': datetime.datetime.utcnow(),
            'lastUse': None,
            'tokenDuration': days,
            'name': name,
            'scope': scope,
            'userId': user['_id'],
            'key': genToken(40),
            'active': active
        }

        return self.setUserAccess(apiKey, user, level=AccessType.ADMIN, save=True)

    def createToken(self, key, days=None):
        """
        Create a token using an API key.

        :param key: The API key (the key itself, not the full document).
        :type key: str
        :param days: You may request a token duration up to the token duration
            of the API key itself, or pass None to use the API key duration.
        :type days: float or None
        """
        from .setting import Setting
        from .token import Token
        from .user import User

        apiKey = self.findOne({
            'key': key
        })

        if apiKey is None or not apiKey['active']:
            raise ValidationException('Invalid API key.')

        cap = apiKey['tokenDuration'] or Setting().get(SettingKey.COOKIE_LIFETIME)
        days = min(float(days or cap), cap)

        user = User().load(apiKey['userId'], force=True)

        # Mark last used stamp
        apiKey['lastUse'] = datetime.datetime.utcnow()
        apiKey = self.save(apiKey)
        token = Token().createToken(user=user, days=days, scope=apiKey['scope'], apiKey=apiKey)
        return (user, token)



# -*- coding: utf-8 -*-
"""This module contains utility methods for parsing girder path strings."""

import re
from ..constants import AccessType
from ..exceptions import AccessException, GirderException, ValidationException
from ..exceptions import ResourcePathNotFound
from .model_importer import ModelImporter
from girder.models.collection import Collection
from girder.models.user import User


# Expose the ResourcePathNotFound exception as its original name
NotFoundException = ResourcePathNotFound


def encode(token):
    """Escape special characters in a token for path representation.

    :param str token: The token to encode
    :return: The encoded string
    :rtype: str
    """
    return token.replace('\\', '\\\\').replace('/', '\\-')


def decode(token):
    """Un-escape special characters in a token from a path representation.

    :param str token: The token to decode
    :return: The decoded string
    :rtype: str
    """
    return token.replace('\\-', '/').replace('\\\\', '\\')


def split(path):
    """Split an encoded path string into decoded tokens.

    :param str path: An encoded path string
    :return: A list of decoded tokens
    :rtype: `list`
    """
    # It would be better to split by the regex `(?<!\\)(?>\\\\)*/`,
    # but python does't support atomic grouping. :(
    chunks = path.split('/')
    processed = [chunks[0]]

    # matches an odd number of backslashes at the end of the string
    escape = re.compile(r'(?<!\\)(?:\\\\)*\\$')

    # Loop through the chunks and check if any of the forward slashes was
    # escaped.
    for chunk in chunks[1:]:
        if escape.search(processed[-1]):
            # join the chunks
            processed[-1] = processed[-1] + '/' + chunk
        else:
            # append a new token
            processed.append(chunk)

    # now decode all of the tokens and return
    return [decode(token) for token in processed]


def join(tokens):
    """Join a list of tokens into an encoded path string.

    :param tokens: A list of tokens
    :return: The encoded path string
    :rtype: str
    """
    return '/'.join([encode(token) for token in tokens])


def lookUpToken(token, parentType, parent):
    """
    Find a particular child resource by name or throw an exception.

    :param token: the name of the child resource to find
    :param parentType: the type of the parent to search
    :param parent: the parent resource
    :returns: the child resource
    """
    # (model name, mask, search filter)
    searchTable = (
        ('folder', parentType in ('user', 'collection', 'folder'), {
            'name': token,
            'parentId': parent['_id'],
            'parentCollection': parentType
        }),
        ('item', parentType == 'folder', {'name': token, 'folderId': parent['_id']}),
        ('file', parentType == 'item', {'name': token, 'itemId': parent['_id']}),
    )

    for candidateModel, mask, filterObject in searchTable:
        if not mask:
            continue

        candidateChild = ModelImporter.model(candidateModel).findOne(filterObject)
        if candidateChild is not None:
            return candidateChild, candidateModel

    # if no folder, item, or file matches, give up
    raise ResourcePathNotFound('Child resource not found: %s(%s)->%s' % (
        parentType, parent.get('name', parent.get('_id')), token))


def lookUpPath(path, user=None, filter=True, force=False):
    """
    Look up a resource in the data hierarchy by path.

    :param path: path of the resource
    :param user: user with correct privileges to access path
    :param filter: Whether the returned model should be filtered.
    :type filter: bool
    :param force: if True, don't validate the access.
    :type force: bool
    """
    path = path.lstrip('/')
    pathArray = split(path)
    model = pathArray[0]

    if model == 'user':
        username = pathArray[1]
        parent = User().findOne({'login': username})

        if parent is None:
            raise ResourcePathNotFound('User not found: %s' % username)

    elif model == 'collection':
        collectionName = pathArray[1]
        parent = Collection().findOne({'name': collectionName})

        if parent is None:
            raise ResourcePathNotFound('Collection not found: %s' % collectionName)

    else:
        raise ValidationException('Invalid path format')

    try:
        document = parent
        if not force:
            ModelImporter.model(model).requireAccess(document, user)
        for token in pathArray[2:]:
            document, model = lookUpToken(token, model, document)
            if not force:
                ModelImporter.model(model).requireAccess(document, user)
    except (ValidationException, AccessException):
        # We should not distinguish the response between access and validation errors so that
        # adversarial users cannot discover the existence of data they don't have access to by
        # looking up a path.
        raise ResourcePathNotFound('Path not found: %s' % path)

    if filter:
        document = ModelImporter.model(model).filter(document, user)

    return {
        'model': model,
        'document': document
    }


def getResourceName(type, doc):
    """
    Get the name of a resource that can be put in a path,

    :param type: the resource model type.
    :type type: str
    :param doc: the resource document.
    :type doc: dict
    :return: the name of the resource.
    :rtype: str
    """
    if type == 'user':
        return doc['login']
    elif type in ('file', 'item', 'folder', 'user', 'collection'):
        return doc['name']
    else:
        raise GirderException('Invalid resource type.')


def getResourcePath(type, doc, user=None, force=False):
    """
    Get the path for a resource.

    :param type: the resource model type.
    :type type: str
    :param doc: the resource document.
    :type doc: dict
    :param user: user with correct privileges to access path
    :type user: dict or None
    :param force: if True, don't validate the access.
    :type force: bool
    :return: the path to the resource.
    :rtype: str
    """
    path = []
    while True:
        path.insert(0, getResourceName(type, doc))
        if type == 'file':
            parentModel = 'item'
            parentId = doc['itemId']
        elif type == 'item':
            parentModel = 'folder'
            parentId = doc['folderId']
        elif type == 'folder':
            parentModel = doc['parentCollection']
            parentId = doc['parentId']
        else:
            break
        doc = ModelImporter.model(parentModel).load(
            id=parentId, user=user, level=AccessType.READ, force=force)
        type = parentModel
    path.insert(0, type)
    return '/' + join(path)



import os

import pytest


@pytest.fixture
def virtualenv_path(host):
    return os.path.join(
        host.user().home,
        # Molecule playbook vars can't be passed into Testinfra tests, so hardcode the path
        '.virtualenvs/girder'
    )


@pytest.fixture
def config_path(host):
    return os.path.join(
        host.user().home,
        '.girder/girder.cfg'
    )


def test_girder_package(host, virtualenv_path):
    pip_path = os.path.join(virtualenv_path, 'bin/pip')
    packages = host.pip_package.get_packages(pip_path=pip_path)
    assert 'girder' in packages


def test_girder_web_build(host, virtualenv_path):
    web_file_path = os.path.join(virtualenv_path, 'share/girder/static/built/girder_app.min.js')
    assert host.file(web_file_path).is_file


def test_girder_service(host):
    girder_service = host.service('girder')
    assert girder_service.is_enabled
    assert girder_service.is_running


def test_girder_socket_private(host):
    girder_socket = host.socket('tcp://127.0.0.1:8080')
    assert girder_socket.is_listening


def test_girder_socket_public(host):
    girder_socket = host.socket('tcp://0.0.0.0:8080')
    assert not girder_socket.is_listening


def test_girder_config_exists(host, config_path):
    assert host.file(config_path).exists


@pytest.mark.parametrize('option, value', [
    (r'server\.socket_host', r'"127\.0\.0\.1"'),
    (r'tools\.proxy\.on', r'True'),
    (r'uri', r'"mongodb://mongodb:27017/girder"'),
    (r'mode', r'"production"')
])
def test_girder_config_content(host, config_path, option, value):
    config_line_re = r'^%s *= *%s$' % (option, value)
    assert host.file(config_path).contains(config_line_re)
